"""ATS model examples."""

from umbi.examples.ats.grid import grid_ats_from_string
from umbi.examples.ats.random_game import random_game_ats
from umbi.examples.ats.random_walk import random_walk_ats

__all__ = [
    "grid_ats_from_string",
    "random_game_ats",
    "random_walk_ats",
]
